BrokerXP = LibStub("AceAddon-3.0"):NewAddon("Broker_XP", "AceConsole-3.0", "AceEvent-3.0","AceTimer-3.0")

local L = LibStub("AceLocale-3.0"):GetLocale("Broker_XP", true)

local bxp = BrokerXP
local time = time
local defaults = {
  profile = {
    dataValue = "XPHour",
  }
}

local xptypes = {
          XPHour = L["XP gain rate (session)"],
          XPHourLevel = L["XP gain rate (level)"],
          XP = L["Current XP"],
          XPpcnt = L["Percentage XP"],
          XPtotal = L["XP gained and total for level"],
          LastLevelGainTime = L["Last level gained"],
          PsuedoInstantaneousXPHour = L["Instantaneous XP rate"],
          XPsession = L["XP gained this session"],
          XPtolevel = L["XP until the next level"],
          TimeToLevel = L["The time at current rate until next level"],
          MobKillsToLevel = L["Mob kills at current rate until next level"],
          LastKill = L["Last kill XP and time"],
          RestExpiryETA = L["The time at current rate until rest runs out"],
          RestXP = L["Amount of rested XP"],
          RestXPpct = L["Percentage of rested XP"],
}

local tooltipordering = {
          "XPtotal",
          "RestXP",
          "XPpcnt",
          "RestXPpct",
          "XPHour",
          "XPsession",
          "XPHourLevel",
          "LastLevelGainTime",
          "PsuedoInstantaneousXPHour",
          "XPtolevel",
          "TimeToLevel",
          "MobKillsToLevel",
          "RestExpiryETA",
          "LastKill",
}

local options = {
    name = L["Broker_XP"],
    handler = bxp,
    type = "group",
  	get = function(info) return bxp.db[info[#info]] end,
  	set = function(info, value) bxp.db[info[#info]] = value end,
    args = {
      dataValue = {
        name = L["Main Value"],
        desc = L["Select the primary value to show"],
        type = "select",
        values = xptypes,
      },
    },
}

function BrokerXP:OnInitialize()
  self.db_root = LibStub("AceDB-3.0"):New("Broker_XPDB",defaults)
  self.db=self.db_root.profile
	local acreg = LibStub("AceConfigRegistry-3.0")
	acreg:RegisterOptionsTable("Broker_XP", options)
	acreg:RegisterOptionsTable("Broker_XP Profiles", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db_root))
	local acdia = LibStub("AceConfigDialog-3.0")
	acdia:AddToBlizOptions("Broker_XP", L["Broker_XP"])
	acdia:AddToBlizOptions("Broker_XP Profiles", L["Profiles"], "Broker_XP")
	
	self.levelledXP = 0
	self.sessionXP = 0
	self.playedTimeOffset = time()
	self.lastLevelTimeOffset = time()
	self.sessionStartTime = time()
end

local lastXP
local recentXPGains={}

function BrokerXP:OnEnable()
	self:RegisterEvent("PLAYER_XP_UPDATE")
	self:RegisterEvent("PLAYER_LEVEL_UP")
	self:RegisterEvent("TIME_PLAYED_MSG")

  RequestTimePlayed()

	self:ScheduleRepeatingTimer("UpdateText",1)
	
	if not self.sessionStartXP then
		self.sessionStartXP = UnitXP("player")
	end
	lastXP = self.sessionStartXP
  table.insert(recentXPGains,{time=time(),xp=lastXP,xpgain=0})
end

function BrokerXP:OnDisable()
	self:UnregisterEvent("PLAYER_XP_UPDATE")
	self:UnregisterEvent("PLAYER_LEVEL_UP")
	self:UnregisterEvent("TIME_PLAYED_MSG")
end

function BrokerXP:PLAYER_XP_UPDATE()
	local newXP = UnitXP("player")
  local lastMobXP = newXP - lastXP
	lastXP = newXP
	if lastMobXP < 0 then
		lastMobXP = 0
	end
	self.sessionXP = newXP - self.sessionStartXP + self.levelledXP
	if lastMobXP > 0 then
	  table.insert(recentXPGains,{time=time(),xp=newXP,xpgain=lastMobXP,rest=GetXPExhaustion()})
	end
end

function BrokerXP:PLAYER_LEVEL_UP()
	self.levelledXP = self.levelledXP + UnitXPMax("player") - self.sessionStartXP
	self.sessionStartXP = 0
	self.lastLevelTimeOffset = time()
end

function BrokerXP:TIME_PLAYED_MSG()
	self.playedTimeOffset = time() - arg1
	self.lastLevelTimeOffset = time() - arg2
end

function BrokerXP:UpdateText()
  local typ=self.db.dataValue or "XPHour"
  local fmt,unit,value,value2,value3=self['meth_'..typ](self)
  if unit then
    self.obj.value=value
    self.obj.unit=unit
  else
    self.obj.value=nil
    self.obj.unit=nil
  end
  self.obj.text=format(fmt,value,value2,value3)
end

function BrokerXP:meth_XP()
  return "%d XP", "XP", UnitXP("player")
end

function BrokerXP:meth_XPpcnt()
  return "%.2f %%", "%", (UnitXP("player")/UnitXPMax("player")) *100
end

function BrokerXP:meth_XPtotal()
  local xp,max=UnitXP("player"),UnitXPMax("player")
  return "%d/%d XP",nil,xp,max
end

function BrokerXP:meth_XPHour()
  return "%.1f XP/hour","XP/hour",self.sessionXP / (time() - self.sessionStartTime) * 3600
end

function BrokerXP:meth_XPHourLevel()
  return "%.1f XP/hour","XP/hour",UnitXP("player") / (time() - self.lastLevelTimeOffset) * 3600
end

function BrokerXP:meth_LastLevelGainTime()
  return "%.1f hours ago", "hours ago", (time() - self.lastLevelTimeOffset) / 3600
end

function BrokerXP:meth_XPsession()
  return "%d XP","XP", self.sessionXP
end

function BrokerXP:meth_XPtolevel()
  return "%d XP","XP", UnitXPMax("player")-UnitXP("player")
end

function BrokerXP:meth_LastKill()
  local last=recentXPGains[#recentXPGains]
  return "%d XP %d seconds ago", nil, last.xpgain,time()-last.time
end

function BrokerXP:meth_RestXP()
  return "%d XP","XP",GetXPExhaustion() or 0
end

function BrokerXP:meth_RestXPpct()
  return "%.1f %%","%",100*(GetXPExhaustion() or 0)/UnitXPMax("player")
end

local function computeXP(history)
  local timespan=0.00000001
  local xpgain=0
  local now=time()
  local currXP=UnitXP("player")
  while #recentXPGains>1 and recentXPGains[1].time+history<now do
    table.remove(recentXPGains,1)
  end
  if #recentXPGains>1 then
    oldest=recentXPGains[1]
    timespan=now-oldest.time
    xpgain=currXP-oldest.xp
  end
  local currentRate=3600*(xpgain/timespan)
  if #recentXPGains>2 then
    local penultimate=recentXPGains[#recentXPGains-1]
    timespan=now-penultimate.time
    xpgain=currXP-penultimate.xp
  end
  local trendRate=3600*(xpgain/timespan)
  return currentRate,trendRate-currentRate,xpgain,#recentXPGains-1
end

function BrokerXP:meth_TimeToLevel()
  local xprate,trend,_,cnt=computeXP(300)
  if cnt>0 then
    local trendstring="no trend"
    if trend>0 then
      trendstring="trending quicker"
    elseif trend<0 then
      trendstring="trending slower"
    end
    local deficit=UnitXPMax("player")-UnitXP("player")
    local hr,mn=math.modf(deficit/xprate)
    return "%dh %dm (%s)",nil,hr,math.floor(mn*60),trendstring
  else
    return "Insufficient data",nil,0
  end
end

function BrokerXP:meth_MobKillsToLevel()
  local _,trend,xpgain,cnt=computeXP(300)
  if cnt>0 then
    local trendstring="no trend"
    if trend>0 then
      trendstring="trending less"
    elseif trend<0 then
      trendstring="trending more"
    end
    local deficit=UnitXPMax("player")-UnitXP("player")
    local mobKills=math.ceil((cnt*deficit)/xpgain)
    return "%d mobs (%s)","mobs",mobKills,trendstring
  else
    return "Insufficient data",nil,0
  end
end
  
function BrokerXP:meth_PsuedoInstantaneousXPHour()
  local rate,trend=computeXP(300)
  local trendstring="0"
  if trend>0 then
    trendstring="+"
  elseif trend<0 then
    trendstring="-"
  end
  return "%.1f XP/hour (%s)","XP/hour",rate,trendstring
end

function BrokerXP:meth_RestExpiryETA()
  local xprate,trend,_,cnt=computeXP(300)
  if cnt>0 then
    local trendstring="no trend"
    if trend>0 then
      trendstring="trending quicker"
    elseif trend<0 then
      trendstring="trending slower"
    end
    local deficit=GetXPExhaustion() or nil
    local hr,mn=math.modf(deficit/xprate)
    return "%dh %dm (%s)",nil,hr,math.floor(mn*60),trendstring
  else
    return "Insufficient data",nil,0
  end
end


function BrokerXP:Tooltip(tooltip)
  tooltip:AddLine(L["Broker_XP"])
  local curr=self.db.dataValue or "XPHour"
  for _,k in pairs(tooltipordering) do
    local label=xptypes[k]
    local fmt,_,value,value2,value3=self["meth_"..k](self)
    local linefmt="%s: "..fmt
    if k==curr then
      linefmt="|cFFFFFFFF"..linefmt
    end
    tooltip:AddLine(format(linefmt,label,value,value2,value3))
  end
end

local ldb = LibStub:GetLibrary("LibDataBroker-1.1", true)
if ldb then 
	BrokerXP.obj=ldb:NewDataObject("Broker_XP", {
		type = "data source",
		text = "",
		OnClick = function(self, button)
			InterfaceOptionsFrame_OpenToCategory(L["Broker_XP"])
		end,
		OnTooltipShow = function(tooltip)
		  BrokerXP:Tooltip(tooltip)
		end,
	})
end

